/*
 * encoder.c
 *
 *  Created on: 16 avr. 2023
 *      Author: Flax
 *
 *      Prefix : ENC
 */

// ==============
// Local includes
// ==============

#include "encoder.h"
#include "register.h"

// ==============
// Local typedefs
// ==============

typedef enum {
  ENC_STM_IDLE,
  ENC_STM_SK01,
  ENC_STM_SK10,
  ENC_STM_SK00,
  ENC_STM_CW01,
  ENC_STM_CW00,
  ENC_STM_CW10,
  ENC_STM_CCW10,
  ENC_STM_CCW00,
  ENC_STM_CCW01,
}e_encoderStm;

typedef struct {
  e_encoderStm nextState_E;
  int8_t increment_S8;
}t_encoderState;

// ===============
// Local constants
// ===============

#define cENCClicLongLength		200 	// Long clic length in number of cyclic task cycles
#define cENCSwActiveState		false 	// Encoder switch active state
#define cENCClicFilter			10 		// Encoder switch debounce filter
#define cENCMoveFilter			1		// Encoder rotation debounce filter

#define cENCEventQueueLength    8		// Event queue length

/*
  Definition of encoder state machine.
  A const tab is defined, with as many elements as the state machine enum.
  For each element, 4 states are defined, depending on the inputs (2 inputs -> 4 combinations).
  https://www.eevblog.com/forum/projects/easy-rotary-encoder-implementation-good-easy-to-implement-code
*/
const t_encoderState cSTMmachine_TA [ENC_STM_CCW01+1][4] =
{
/* 00                 01                  10                  11  = INPUTS*/
  {{ENC_STM_SK00, 0}, {ENC_STM_CW01, 0},  {ENC_STM_CCW10, 0}, {ENC_STM_IDLE, 0}},   // ENC_STM_IDLE
  {{ENC_STM_CW00, 0}, {ENC_STM_SK01, 0},  {ENC_STM_CCW10, 0}, {ENC_STM_IDLE, -1}},  // ENC_STM_SK01
  {{ENC_STM_CCW00, 0}, {ENC_STM_CW01, 0}, {ENC_STM_SK10, 0},  {ENC_STM_IDLE, 1}},   // ENC_STM_SK10
  {{ENC_STM_SK00, 0}, {ENC_STM_CCW01, 0}, {ENC_STM_CW10, 0},  {ENC_STM_IDLE, 0}},   // ENC_STM_SK00
  {{ENC_STM_CW00, 0}, {ENC_STM_CW01, 0},  {ENC_STM_SK10, 0},  {ENC_STM_IDLE, 0}},   // ENC_STM_CW01
  {{ENC_STM_CW00, 0}, {ENC_STM_CW01, 0},  {ENC_STM_CW10, 0},  {ENC_STM_IDLE, 1}},   // ENC_STM_CW00
  {{ENC_STM_CW00, 0}, {ENC_STM_CW01, 1},  {ENC_STM_CW10, 0},  {ENC_STM_IDLE, 1}},   // ENC_STM_CW10
  {{ENC_STM_CCW00, 0}, {ENC_STM_SK01, 0}, {ENC_STM_CCW10, 0}, {ENC_STM_IDLE, 0}},   // ENC_STM_CCW10
  {{ENC_STM_CCW00, 0}, {ENC_STM_CCW01, 0}, {ENC_STM_CCW10, 0}, {ENC_STM_IDLE, -1}},   // ENC_STM_CCW00
  {{ENC_STM_CCW00, 0}, {ENC_STM_CCW01, 0}, {ENC_STM_CCW10, -1}, {ENC_STM_IDLE, -1}},   // ENC_STM_CCW01
};

// ==========================
// Local functions prototypes
// ==========================

static bool ENCPushEvent (e_ENCEventType *evt);

// ===========================
// Local functions definitions
// ===========================

// ===============
// Local variables
// ===============

static e_ENCEventType encoder_event_prev_E;
static e_ENCEventType encoder_event_current_E;
static uint16_t encoder_clic_long_cnt_u16; // Counter for long clic detection
static uint16_t encoder_clic_cnt_u16; // Counter for clic filtering

// Event queue
static e_ENCEventType ENCEventQueue[cENCEventQueueLength];
static uint8_t ENCEventQueueCurrentIndex_U8;
static uint8_t ENCEventQueueWriteIndex_U8;

static e_encoderStm STMnextState_E;
static int8_t increment_S8;

static e_ENCEventType ENCEvent_E; // Local current event
static bool ENCNewEventAvailable_B; // Indicates if the event queue contains new events

// ================
// Shared variables
// ================

// ============================
// Shared functions definitions
// ============================

void ENCInterrupt (void)
/* Interrupt routine
 * Updates the encoder state depending on input events.
 * Must be called on each edge of each encoder input.
 * */
{
	uint8_t state_U8 = 0U;
	// Read ENC1 state, increment state by 1
	if (mREGInputsENC1GetState())
	{
		state_U8 += 1U;
	}

	// Read ENC2 state, increment state by 2
	if (mREGInputsENC2GetState())
	{
		state_U8 += 2U;
	}

	increment_S8 += cSTMmachine_TA[STMnextState_E][state_U8].increment_S8;
	STMnextState_E = cSTMmachine_TA[STMnextState_E][state_U8].nextState_E;

	if (increment_S8 > 0)
	{
		ENCEvent_E = ENC_EVENT_INC;
		ENCPushEvent(&ENCEvent_E);
		ENCEvent_E = ENC_EVENT_NONE;
		increment_S8 = 0;
	}
	else if (increment_S8 < 0)
	{
		ENCEvent_E = ENC_EVENT_DEC;
		ENCPushEvent(&ENCEvent_E);
		ENCEvent_E = ENC_EVENT_NONE;
		increment_S8 = 0;
	}
}

void ENCInit(void)
{
	ENCEventQueueCurrentIndex_U8 = 0U;
	ENCEventQueueWriteIndex_U8 = 0U;

	STMnextState_E = ENC_STM_IDLE;
	increment_S8 = 0;
}

void ENCStart(void)
{
	encoder_event_current_E = ENC_EVENT_NONE;
	encoder_event_prev_E = ENC_EVENT_NONE;
	encoder_clic_cnt_u16 = 0U;
	encoder_clic_long_cnt_u16 = 0U;

	// ENC1 : EXTI0
	mREGInputsENC1ExtiEnableIT(); // Enable ENC1 interrrupt
	// ENC2 : EXTI3
	mREGInputsENC2ExtiEnableIT(); // Enable ENC2 interrrupt
}

void ENCStop(void)
{

}

void ENCCyclicTask(void)
/* Cyclic treatment
 * Must be called at a rate suitable for HMI reactivity.
 * */
{
	uint8_t state_U8 = 0U;
	// Read ENC1 state, increment state by 1
	if (mREGInputsENC1GetState())
	{
		state_U8 += 1U;
	}

	// Read ENC2 state, increment state by 2
	if (mREGInputsENC2GetState())
	{
		state_U8 += 2U;
	}

	increment_S8 += cSTMmachine_TA[STMnextState_E][state_U8].increment_S8;
	STMnextState_E = cSTMmachine_TA[STMnextState_E][state_U8].nextState_E;

	if (increment_S8 > 0)
	{
		ENCEvent_E = ENC_EVENT_INC;
		ENCPushEvent(&ENCEvent_E);
		ENCEvent_E = ENC_EVENT_NONE;
		increment_S8 = 0;
	}
	else if (increment_S8 < 0)
	{
		ENCEvent_E = ENC_EVENT_DEC;
		ENCPushEvent(&ENCEvent_E);
		ENCEvent_E = ENC_EVENT_NONE;
		increment_S8 = 0;
	}

	if (mREGInputsSWGetState())
	{ // Clic event detected
		encoder_event_current_E = ENC_EVENT_CLIC;
	}
	else if (encoder_event_prev_E == ENC_EVENT_CLIC_LONG)
	{ // If previous detected event was a long clic -> the SW is being maintained -> ignore
		ENCEvent_E = ENC_EVENT_NONE;
	}

	if ((encoder_event_current_E == ENC_EVENT_CLIC) && (encoder_clic_long_cnt_u16 >= (uint16_t)cENCClicLongLength))
	{ // SW is being pressed since at least the long clic timeout -> long clic event detected
		encoder_event_current_E = ENC_EVENT_CLIC_LONG;
	}

	switch (encoder_event_current_E)
	{
	case ENC_EVENT_NONE :
		if((encoder_event_prev_E == ENC_EVENT_CLIC) &&
				(encoder_clic_long_cnt_u16 < (uint16_t)cENCClicLongLength) &&
				(encoder_clic_cnt_u16 >= cENCClicFilter))
		{ // If the previous event was a short clic event -> falling edge on a short clic
			ENCEvent_E = ENC_EVENT_CLIC;
			ENCPushEvent(&ENCEvent_E);
			ENCEvent_E = ENC_EVENT_NONE;
			encoder_clic_cnt_u16 = 0U;
		}
		encoder_clic_long_cnt_u16 = 0U;
		break;

	case ENC_EVENT_CLIC :
		if ((encoder_event_prev_E == ENC_EVENT_CLIC) && (encoder_clic_long_cnt_u16 < (uint16_t)cENCClicLongLength))
		{ // If the switch is being maintained -> increment the clic timeout counter
			encoder_clic_long_cnt_u16++;
		}
		break;

	case ENC_EVENT_CLIC_LONG :
		// Long clic is triggered once the long clic delay, does not wait for a switch release
		ENCEvent_E = ENC_EVENT_CLIC_LONG;
		if (encoder_event_prev_E != ENC_EVENT_CLIC_LONG)
		{ // One-time event protection
			ENCPushEvent(&ENCEvent_E);
		}
		break;

	default :
		encoder_clic_long_cnt_u16 = 0U;
		break;
	}

	// Clic filter counter
	if (encoder_clic_cnt_u16 < cENCClicFilter)
	{
		encoder_clic_cnt_u16++;
	}

	encoder_event_prev_E = encoder_event_current_E; // Memorise previous event
	encoder_event_current_E = ENC_EVENT_NONE; // Reset current event
}

bool ENCPushEvent (e_ENCEventType *evt)
/* Push an event in the event queue
 * */
{
	ENCEventQueueWriteIndex_U8++;
	ENCNewEventAvailable_B = true;

	if (ENCEventQueueWriteIndex_U8 == ENCEventQueueCurrentIndex_U8)
	{ // Queue overflow
		return (false);
	}

	if (ENCEventQueueWriteIndex_U8 >= cENCEventQueueLength)
	{
		ENCEventQueueWriteIndex_U8 = 0U;
	}

	ENCEventQueue[ENCEventQueueWriteIndex_U8] = *evt;

	return (true);
}

bool ENCPopEvent (e_ENCEventType *evt)
/* Pop an event from the event queue
 * */
{
	bool ret_B = false;

	if (ENCNewEventAvailable_B == true)
	{ // If new events are available, return true
		ret_B = true;

		// Get the next event in the queue
		if (ENCEventQueueCurrentIndex_U8 != ENCEventQueueWriteIndex_U8)
		{
			ENCEventQueueCurrentIndex_U8++;
			if (ENCEventQueueCurrentIndex_U8 >= cENCEventQueueLength)
			{
				ENCEventQueueCurrentIndex_U8 = 0U;
			}
		}
		else
		{
			ENCNewEventAvailable_B = false;
		}
	} // If no new event is available, return false

	// Provide the event that has been found
	// If this event is not new, then a NONE event will be provided
	*evt = ENCEventQueue[ENCEventQueueCurrentIndex_U8];
	// Reset the event - One-time read
	ENCEventQueue[ENCEventQueueCurrentIndex_U8] = ENC_EVENT_NONE;

	return (ret_B);
}

bool ENCIsEventAvailable (void)
/* Indicates if a new event is available on the event queue
 * */
{
	return (ENCNewEventAvailable_B);
}
