/*
 * hmi.c
 *
 *  Created on: Apr 16, 2023
 *      Author: Flax
 *
 *      Prefix : HMI
 */

// ==============
// Local includes
// ==============
#include "hmi.h"
#include "stdint.h"
#include "encoder.h"
#include "temperatureRegulation.h"
#include "display.h"
#include "register.h"

// ==============
// Local typedefs
// ==============
typedef enum {
	HMI_DISPLAY_MODE_RESET,
	HMI_DISPLAY_MODE_OFF,
	HMI_DISPLAY_MODE_TEMP,
	HMI_DISPLAY_MODE_ERR,
	HMI_DISPLAY_MODE_STANDBY,
	HMI_DISPLAY_MODE_PARAM_TEMP_SPT_TITLE,
	HMI_DISPLAY_MODE_PARAM_TEMP_SPT,
	HMI_DISPLAY_MODE_PARAM_TEMP_STBY_TITLE,
	HMI_DISPLAY_MODE_PARAM_TEMP_STBY,
	HMI_DISPLAY_MODE_PARAM_GAIN_TITLE,
	HMI_DISPLAY_MODE_PARAM_GAIN,
	HMI_DISPLAY_MODE_MAX
}e_HMIDisplayModeType;

// ===============
// Local constants
// ===============
#define cHMIInputEventQueueLength    8		// Event queue length

#define cHMIClicLongLength		400 	// Long clic length in number of cyclic task cycles
#define cHMIClicFilter			10 		// Encoder switch debounce filter

#define cHMIDisplayParamTime	2000	// Timeout during which the parameter is shown before going back to previous mode
#define cHMIDisplayTitleTime	1000	// Timeout during which the parameter title is shown before showing the parameter

// ==========================
// Local functions prototypes
// ==========================
static void HmiGetInputs (void);
static void HmiUpdateDisplay (void);
static void HmiSetOutputs (void);

static bool HmiPushInputEvent (e_HMIInputEventType *evt);

// ===============
// Local variables
// ===============
static e_HMIInputEventType input_event_prev_E;
static e_HMIInputEventType input_event_current_E;
static uint16_t sw_clic_long_cnt_u16; // Counter for long clic detection
static uint16_t sw_clic_cnt_u16; // Counter for clic filtering
static bool sw_state_B;
static bool sw_state_prev_B;

// Event queue
static e_HMIInputEventType HMIInputEventQueue[cHMIInputEventQueueLength];
static uint8_t HMIInputEventQueueCurrentIndex_U8;
static uint8_t HMIInputEventQueueWriteIndex_U8;
static bool HMINewInputEventAvailable_B;

// Display
static e_HMIDisplayModeType display_mode_current_E;
static e_HMIDisplayModeType display_mode_prev_E;
static uint16_t display_param_cnt_u16;

// DEBUG
static uint16_t temperature_debug;
static uint16_t temperature_stb_debug;
static uint16_t gain_debug;

// ===========================
// Local functions definitions
// ===========================
static void HmiGetInputs (void)
{
	e_ENCEventType encoder_evt;
	e_HMIInputEventType input_evt;

	// Check encoder events
	if (ENCPopEvent(&encoder_evt))
	{
		switch (encoder_evt)
		{
		case ENC_EVENT_NONE:
			break;
		case ENC_EVENT_INC:
			input_evt = HMI_EVENT_INPUT_MOVE_RIGHT;
			HmiPushInputEvent(&input_evt);
			break;
		case ENC_EVENT_DEC:
			input_evt = HMI_EVENT_INPUT_MOVE_LEFT;
			HmiPushInputEvent(&input_evt);
			break;
		default:
			break;
		}
	}

	// Check encoder button
	// --------------------
	sw_state_B = mREGInputsSWGetState();

	// Identify "local" action
	if (sw_state_B)
	{ // Button is pushed
		// Increment debounce counter
		if (sw_clic_cnt_u16 < cHMIClicFilter)
		{
			sw_clic_cnt_u16++;
		}
		else
		{ // Debounce filter elapsed
			input_event_current_E = HMI_EVENT_INPUT_CLIC_SHORT;
		}

		if (sw_clic_long_cnt_u16 < cHMIClicLongLength)
		{
			sw_clic_long_cnt_u16++;
		}
		else
		{ // Long clic timer elapsed
			input_event_current_E = HMI_EVENT_INPUT_CLIC_LONG;
		}
	}
	else
	{
		input_event_current_E = HMI_EVENT_INPUT_NONE;

		sw_clic_cnt_u16 = 0U;
		sw_clic_long_cnt_u16 = 0U;
	}

	// Identify functional HMI action
	switch (input_event_current_E)
	{
	case HMI_EVENT_INPUT_NONE:
		if(input_event_prev_E == HMI_EVENT_INPUT_CLIC_SHORT)
		{ // Short clic detected - Triggered upon button release
			HmiPushInputEvent(&input_event_prev_E); // Push short clic
		}
		break;
	case HMI_EVENT_INPUT_CLIC_SHORT:
		break;
	case HMI_EVENT_INPUT_CLIC_LONG:
		if (input_event_prev_E != HMI_EVENT_INPUT_CLIC_LONG)
		{ // Long clic detected - Triggered once, upon timer elapsed
			HmiPushInputEvent(&input_event_current_E); // Push long clic
		}
		break;
	default:
		break;
	}

	// Update previous switch state
	sw_state_prev_B = sw_state_B;
	input_event_prev_E = input_event_current_E;
}

static void HmiUpdateDisplay (void)
{
	e_HMIInputEventType evt;

	// Display main state machine
	switch(display_mode_current_E)
	{

	/*** Reset state ***/
	/*******************/
	case HMI_DISPLAY_MODE_RESET:
		display_mode_current_E = HMI_DISPLAY_MODE_STANDBY; // First state after reset is standby
		break;

	/*** OFF state ***/
	/*****************/
	case HMI_DISPLAY_MODE_OFF:
		DISPUpdateDigit(0, DISP_CHAR_7SEG_O);
		DISPUpdateDigit(1, DISP_CHAR_7SEG_F);
		DISPUpdateDigit(2, DISP_CHAR_7SEG_F);

		if (HMIPopInputEvent(&evt))
		{ // Get next input event
			switch (evt)
			{
			case HMI_EVENT_INPUT_MOVE_RIGHT:
			case HMI_EVENT_INPUT_MOVE_LEFT:
				break;
			case HMI_EVENT_INPUT_CLIC_SHORT:
				// Edit parameters
				display_mode_current_E = HMI_DISPLAY_MODE_PARAM_TEMP_SPT_TITLE;
				break;
			case HMI_EVENT_INPUT_CLIC_LONG:
				// Go to normal mode
				display_mode_current_E = HMI_DISPLAY_MODE_TEMP;
				break;
			default:
				break;
			}
		}
		break;

	/*** Display temperature set point title ***/
	/*******************************************/
	case HMI_DISPLAY_MODE_PARAM_TEMP_SPT_TITLE:
		DISPUpdateDigit(0, DISP_CHAR_7SEG_S);
		DISPUpdateDigit(1, DISP_CHAR_7SEG_P);
		DISPUpdateDigit(2, DISP_CHAR_7SEG_T);
		if (display_param_cnt_u16 > 0)
		{
			display_param_cnt_u16--;
		}
		else
		{ // Title display elapsed, go to actual param edit
			display_param_cnt_u16 = cHMIDisplayParamTime;
			display_mode_current_E = HMI_DISPLAY_MODE_PARAM_TEMP_SPT;
		}
		break;

	/*** Display temperature set point ***/
	/*************************************/
	case HMI_DISPLAY_MODE_PARAM_TEMP_SPT:
		// Display set point temperature
		DISPUpdateValueNumber(REGSPTGet());

		if (display_param_cnt_u16 > 0)
		{
			display_param_cnt_u16--;
		}
		else
		{
			if (REGIsOff())
			{
				// Do nothing
			}
			else if (REGIsStandby())
			{ // Go back to standby
				display_mode_current_E = HMI_DISPLAY_MODE_STANDBY;
			}
			else
			{ // Go back to normal
				display_mode_current_E = HMI_DISPLAY_MODE_TEMP;
			}
		}

		// Check for input events
		if (HMIPopInputEvent(&evt))
		{ // Get input event
			switch (evt)
			{
			case HMI_EVENT_INPUT_MOVE_RIGHT:
				REGSPTSet(REGSPTGet() + 1);
				// Reset display timer
				display_param_cnt_u16 = cHMIDisplayParamTime;
				break;
			case HMI_EVENT_INPUT_MOVE_LEFT:
				REGSPTSet(REGSPTGet() - 1);
				// Reset display timer
				display_param_cnt_u16 = cHMIDisplayParamTime;
				break;
			case HMI_EVENT_INPUT_CLIC_SHORT:
				if (REGIsOff())
				{
					display_param_cnt_u16 = cHMIDisplayTitleTime;
					display_mode_current_E = HMI_DISPLAY_MODE_PARAM_TEMP_STBY_TITLE;
				}
				else if (REGIsStandby())
				{
					display_mode_current_E = HMI_DISPLAY_MODE_STANDBY;
				}
				else
				{
					display_mode_current_E = HMI_DISPLAY_MODE_TEMP;
				}
				break;
			case HMI_EVENT_INPUT_CLIC_LONG:
				// Cancel
				display_mode_current_E = HMI_DISPLAY_MODE_OFF;
				break;
			default:
				break;
			}
		}
		break;

	/*** Display standby temperature set point title ***/
	/***************************************************/
	case HMI_DISPLAY_MODE_PARAM_TEMP_STBY_TITLE:
		DISPUpdateDigit(0, DISP_CHAR_7SEG_S);
		DISPUpdateDigit(1, DISP_CHAR_7SEG_P);
		DISPUpdateDigit(2, DISP_CHAR_7SEG_S);
		if (display_param_cnt_u16 > 0)
		{
			display_param_cnt_u16--;
		}
		else
		{ // Title display elapsed, go to actual param edit
			display_param_cnt_u16 = cHMIDisplayParamTime;
			display_mode_current_E = HMI_DISPLAY_MODE_PARAM_TEMP_STBY;
		}
		break;

	/*** Display standby temperature set point ***/
	/*********************************************/
	case HMI_DISPLAY_MODE_PARAM_TEMP_STBY:
		// Display set point temperature
		DISPUpdateValueNumber(REGSPTStbyGet());

		if (display_param_cnt_u16 > 0)
		{
			display_param_cnt_u16--;
		}
		else
		{
			if (REGIsOff())
			{
				// Do nothing
			}
			else if (REGIsStandby())
			{ // Go back to standby
				display_mode_current_E = HMI_DISPLAY_MODE_STANDBY;
			}
			else
			{ // Go back to normal
				display_mode_current_E = HMI_DISPLAY_MODE_TEMP;
			}
		}

		// Check for input events
		if (HMIPopInputEvent(&evt))
		{ // Get next input event
			switch (evt)
			{
			case HMI_EVENT_INPUT_MOVE_RIGHT:
				REGSPTStbySet(REGSPTStbyGet() + 1);
				// Reset display timer
				display_param_cnt_u16 = cHMIDisplayParamTime;
				break;
			case HMI_EVENT_INPUT_MOVE_LEFT:
				REGSPTStbySet(REGSPTStbyGet() - 1);
				// Reset display timer
				display_param_cnt_u16 = cHMIDisplayParamTime;
				break;
			case HMI_EVENT_INPUT_CLIC_SHORT:
				if (REGIsOff())
				{
					display_mode_current_E = HMI_DISPLAY_MODE_PARAM_GAIN_TITLE;
				}
				else if (REGIsStandby())
				{
					display_mode_current_E = HMI_DISPLAY_MODE_STANDBY;
				}
				else
				{
					display_mode_current_E = HMI_DISPLAY_MODE_TEMP;
				}
				break;
			case HMI_EVENT_INPUT_CLIC_LONG:
				// Cancel
				display_mode_current_E = HMI_DISPLAY_MODE_OFF;
				break;
			default:
				break;
			}
		}
		break;

	/*** Display gain title ***/
	/**************************/
	case HMI_DISPLAY_MODE_PARAM_GAIN_TITLE:
		DISPUpdateDigit(0, DISP_CHAR_7SEG_G);
		DISPUpdateDigit(1, DISP_CHAR_7SEG_A);
		DISPUpdateDigit(2, DISP_CHAR_7SEG_N);

		if (display_param_cnt_u16 > 0)
		{
			display_param_cnt_u16--;
		}
		else
		{ // Title display elapsed, go to actual param edit
			display_param_cnt_u16 = cHMIDisplayParamTime;
			display_mode_current_E = HMI_DISPLAY_MODE_PARAM_GAIN;
		}
		break;

	/*** Display gain value ***/
	/**************************/
	case HMI_DISPLAY_MODE_PARAM_GAIN:
		// Check for input events
		if (HMIPopInputEvent(&evt))
		{ // Get next input event
			switch (evt)
			{
			case HMI_EVENT_INPUT_MOVE_RIGHT:
				gain_debug++;
				break;
			case HMI_EVENT_INPUT_MOVE_LEFT:
				gain_debug--;
				break;
			case HMI_EVENT_INPUT_CLIC_SHORT:
				display_mode_current_E = HMI_DISPLAY_MODE_OFF;
				break;
			case HMI_EVENT_INPUT_CLIC_LONG:
				// Cancel
				display_mode_current_E = HMI_DISPLAY_MODE_OFF;
				break;
			default:
				break;
			}
		}
		break;

	/*** Display actual tip temperature ***/
	/**************************************/
	case HMI_DISPLAY_MODE_TEMP:
		// Display actual tip temperature
		DISPUpdateValueNumber(REGTempGet());

		display_param_cnt_u16 = cHMIDisplayTitleTime;
		if (REGIsStandby())
		{
			display_mode_current_E = HMI_DISPLAY_MODE_STANDBY;
		}
		else if (HMIPopInputEvent(&evt))
		{ // Get next input event
			switch (evt)
			{
			case HMI_EVENT_INPUT_MOVE_RIGHT:
			case HMI_EVENT_INPUT_MOVE_LEFT:
				// Encoder movement : go to SPT display
				display_mode_current_E = HMI_DISPLAY_MODE_PARAM_TEMP_SPT_TITLE;
				break;
			case HMI_EVENT_INPUT_CLIC_SHORT:
				// Switch to standby
				REGStandbySet();
				break;
			case HMI_EVENT_INPUT_CLIC_LONG:
				// Switch OFF
				display_mode_current_E = HMI_DISPLAY_MODE_OFF;
				break;
			default:
				break;
			}
		}
		break;

	/*** Display error ***/
	/*********************/
	case HMI_DISPLAY_MODE_ERR:
		DISPUpdateDigit(0, DISP_CHAR_7SEG_E);
		DISPUpdateDigit(1, DISP_CHAR_7SEG_R);
		DISPUpdateDigit(2, DISP_CHAR_7SEG_R);

		if (REGIsError())
		{
			if (HMIPopInputEvent(&evt))
			{ // Get next input event
				if (evt == HMI_EVENT_INPUT_CLIC_LONG)
				{ // Long clic to reset error
					REGErrorReset(); // Try to reset error
				}
			}
		}
		else
		{ // If no error is present, go back to standby
			display_mode_current_E = HMI_DISPLAY_MODE_STANDBY;
		}
		break;

	/*** Display standby ***/
	/***********************/
	case HMI_DISPLAY_MODE_STANDBY:
		DISPUpdateDigit(0, DISP_CHAR_7SEG_S);
		DISPUpdateDigit(1, DISP_CHAR_7SEG_T);
		DISPUpdateDigit(2, DISP_CHAR_7SEG_B);

		if (REGIsStandby())
		{
			if (HMIPopInputEvent(&evt))
			{ // Get next input event
				switch (evt)
				{
				case HMI_EVENT_INPUT_MOVE_RIGHT:
				case HMI_EVENT_INPUT_MOVE_LEFT:
					// Encoder movement : go to SPT display
					display_mode_current_E = HMI_DISPLAY_MODE_PARAM_TEMP_STBY_TITLE;
					break;
				case HMI_EVENT_INPUT_CLIC_SHORT:
					// Switch to normal
					REGStandbyReset();
					break;
				case HMI_EVENT_INPUT_CLIC_LONG:
					REGOffSet();
					break;
				default:
					break;
				}
			}
		}
		else if (REGIsOff())
		{
			display_mode_current_E = HMI_DISPLAY_MODE_OFF;
		}
		else
		{
			display_mode_current_E = HMI_DISPLAY_MODE_TEMP;
		}
		break;

	/*** Default ***/
	/***************/
	default:
		display_mode_current_E = HMI_DISPLAY_MODE_OFF;
		break;
	}

	// Memorise previous state
	display_mode_prev_E = display_mode_current_E;

	// Refresh display
	DISPRefresh();
}

static void HmiSetOutputs (void)
{
	// DISPUpdateValueNumber(temperature_debug);
}

static bool HmiPushInputEvent (e_HMIInputEventType *evt)
/* Push an event in the event queue
 * */
{
	HMIInputEventQueueWriteIndex_U8++;
	HMINewInputEventAvailable_B = true;

	if (HMIInputEventQueueWriteIndex_U8 == HMIInputEventQueueCurrentIndex_U8)
	{ // Queue overflow
		return (false);
	}

	if (HMIInputEventQueueWriteIndex_U8 >= cHMIInputEventQueueLength)
	{
		HMIInputEventQueueWriteIndex_U8 = 0U;
	}

	HMIInputEventQueue[HMIInputEventQueueWriteIndex_U8] = *evt;

	return (true);
}

// ============================
// Shared functions definitions
// ============================

void HMIInit (void)
{
	REGOffSet();
	display_mode_prev_E = HMI_DISPLAY_MODE_RESET;
	display_mode_current_E = HMI_DISPLAY_MODE_RESET;
	display_param_cnt_u16 = cHMIDisplayTitleTime;

	DISPInit();
}

void HMIStart (void)
{
	DISPStart();
}

void HMIStop (void)
{
	DISPStop();
}

void HMICyclicTask (void)
{
	HmiGetInputs();
	HmiUpdateDisplay();
	HmiSetOutputs();
}



bool HMIPopInputEvent (e_HMIInputEventType *evt)
/* Pop an event from the event queue
 * */
{
	bool ret_B = false;

	if (HMINewInputEventAvailable_B == true)
	{ // If new events are available, return true
		ret_B = true;

		// Get the next event in the queue
		if (HMIInputEventQueueCurrentIndex_U8 != HMIInputEventQueueWriteIndex_U8)
		{
			HMIInputEventQueueCurrentIndex_U8++;
			if (HMIInputEventQueueCurrentIndex_U8 >= cHMIInputEventQueueLength)
			{
				HMIInputEventQueueCurrentIndex_U8 = 0U;
			}
		}
		else
		{
			HMINewInputEventAvailable_B = false;
		}
	} // If no new event is available, return false

	// Provide the event that has been found
	// If this event is not new, then a NONE event will be provided
	*evt = HMIInputEventQueue[HMIInputEventQueueCurrentIndex_U8];
	// Reset the event - One-time read
	HMIInputEventQueue[HMIInputEventQueueCurrentIndex_U8] = HMI_EVENT_INPUT_NONE;

	return (ret_B);
}

bool HMIIsInputEventAvailable (void)
/* Indicates if a new event is available on the event queue
 * */
{
	return (HMINewInputEventAvailable_B);
}
